@@ -1,4 +1,4 @@
-# Copyright (c) 2018-2019 Micro Focus or one of its affiliates.
+# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
 # Copyright (c) 2018 Uber Technologies, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -43,7 +43,6 @@ from decimal import Decimal
 from uuid import UUID
 
 # noinspection PyCompatibility,PyUnresolvedReferences
-from builtins import str
 from dateutil import parser, tz
 
 from .. import errors
@@ -138,27 +137,27 @@ def time_parse(s):
 # Type casting of SQL types bytes representation into Python objects
 def vertica_type_cast(type_code, unicode_error):
     typecaster = {
-        VerticaType.UNKNOWN: None,
+        VerticaType.UNKNOWN: bytes,
         VerticaType.BOOL: lambda s: s == b't',
         VerticaType.INT8: lambda s: int(s),
         VerticaType.FLOAT8: lambda s: float(s),
-        VerticaType.CHAR: lambda s: str(s, encoding='utf-8', errors=unicode_error),
-        VerticaType.VARCHAR: lambda s: str(s, encoding='utf-8', errors=unicode_error),
+        VerticaType.CHAR: lambda s: s.decode('utf-8', unicode_error),
+        VerticaType.VARCHAR: lambda s: s.decode('utf-8', unicode_error),
         VerticaType.DATE: date_parse,
         VerticaType.TIME: time_parse,
         VerticaType.TIMESTAMP: timestamp_parse,
         VerticaType.TIMESTAMPTZ: timestamp_tz_parse,
-        VerticaType.INTERVAL: None,
-        VerticaType.TIMETZ: None,
-        VerticaType.NUMERIC: lambda s: Decimal(str(s, encoding='utf-8', errors=unicode_error)),
-        VerticaType.VARBINARY: None,
-        VerticaType.UUID: lambda s: UUID(str(s, encoding='utf-8', errors=unicode_error)),
-        VerticaType.INTERVALYM: None,
-        VerticaType.LONGVARCHAR: lambda s: str(s, encoding='utf-8', errors=unicode_error),
-        VerticaType.LONGVARBINARY: None,
-        VerticaType.BINARY: None
+        VerticaType.INTERVAL: bytes,
+        VerticaType.TIMETZ: bytes,
+        VerticaType.NUMERIC: lambda s: Decimal(s.decode('utf-8', unicode_error)),
+        VerticaType.VARBINARY: bytes,
+        VerticaType.UUID: lambda s: UUID(s.decode('utf-8', unicode_error)),
+        VerticaType.INTERVALYM: bytes,
+        VerticaType.LONGVARCHAR: lambda s: s.decode('utf-8', unicode_error),
+        VerticaType.LONGVARBINARY: bytes,
+        VerticaType.BINARY: bytes
     }
-    return typecaster.get(type_code, None)
+    return typecaster.get(type_code, bytes)
 
 
 ColumnTuple = namedtuple('Column', ['name', 'type_code', 'display_size', 'internal_size',
@@ -184,7 +183,7 @@ class Column(object):
     def convert(self, s):
         if s is None:
             return
-        return self.converter(s) if self.converter is not None else s
+        return self.converter(s)
 
     def __str__(self):
         return as_str(str(self.props))
