@@ -1,4 +1,4 @@
-# Copyright (c) 2018-2019 Micro Focus or one of its affiliates.
+# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
 # Copyright (c) 2018 Uber Technologies, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -46,7 +46,6 @@ from struct import unpack
 from collections import deque, namedtuple
 
 # noinspection PyCompatibility,PyUnresolvedReferences
-from builtins import str
 from six import raise_from, string_types, integer_types, PY2
 
 if PY2:
@@ -98,21 +97,31 @@ def parse_dsn(dsn):
         ('password', url.password),
         ('database', url.path[1:])) if v
     }
-    for key, value in parse_qs(url.query).items():
-        if key == 'backup_server_node':
+    for key, values in parse_qs(url.query, keep_blank_values=True).items():
+        # Try to get the last non-blank value in the list of values for each key
+        for i in reversed(range(len(values))):
+            value = values[i]
+            if value != '':
+                break
+
+        if value == '' and key != 'log_path':
+            # blank values are to be ignored
             continue
-        elif key in ('connection_load_balance', 'use_prepared_statements', 'ssl'):
-            lower = value[-1].lower()
+        elif key == 'backup_server_node':
+            continue
+        elif key in ('connection_load_balance', 'use_prepared_statements',
+                     'disable_copy_local', 'ssl'):
+            lower = value.lower()
             if lower in ('true', 'on', '1'):
                 result[key] = True
             elif lower in ('false', 'off', '0'):
                 result[key] = False
         elif key == 'connection_timeout':
-            result[key] = float(value[-1])
-        elif key == 'log_level' and value[-1].isdigit():
-            result[key] = int(value[-1])
+            result[key] = float(value)
+        elif key == 'log_level' and value.isdigit():
+            result[key] = int(value)
         else:
-            result[key] = value[-1]
+            result[key] = value
 
     return result
 
@@ -243,7 +252,7 @@ class Connection(object):
         options = options or {}
         self.options = parse_dsn(options['dsn']) if 'dsn' in options else {}
         self.options.update({key: value for key, value in options.items() \
-                             if key != 'dsn' and value is not None})
+                             if key == 'log_path' or (key != 'dsn' and value is not None)})
 
         # Set up connection logger
         logger_name = 'vertica_{0}_{1}'.format(id(self), str(uuid.uuid4())) # must be a unique value
@@ -289,6 +298,11 @@ class Connection(object):
         self._logger.debug('Connection prepared statements is {}'.format(
                      'enabled' if self.options['use_prepared_statements'] else 'disabled'))
 
+        # knob for disabling COPY LOCAL operations
+        self.options.setdefault('disable_copy_local', False)
+        self._logger.debug('COPY LOCAL operation is {}'.format(
+                     'disabled' if self.options['disable_copy_local'] else 'enabled'))
+
         self._logger.info('Connecting as user "{}" to database "{}" on host "{}" with port {}'.format(
                      self.options['user'], self.options['database'],
                      self.options['host'], self.options['port']))
@@ -586,6 +600,15 @@ class Connection(object):
             else:
                 self._logger.warning(message.error_message())
 
+    def read_string(self):
+        s = bytearray()
+        while True:
+            char = self.read_bytes(1)
+            if char == b'\x00':
+                break
+            s.extend(char)
+        return s
+
     def read_message(self):
         while True:
             try:
@@ -593,7 +616,27 @@ class Connection(object):
                 size = unpack('!I', self.read_bytes(4))[0]
                 if size < 4:
                     raise errors.MessageError("Bad message size: {0}".format(size))
-                message = BackendMessage.from_type(type_, self.read_bytes(size - 4))
+                if type_ == messages.WriteFile.message_id:
+                    # The whole WriteFile message may not be read at here.
+                    # Instead, only the file name and file length is read.
+                    # This is because the message could be too large to read all at once.
+                    f = self.read_string()
+                    filename = f.decode('utf-8')
+                    file_length = unpack('!I', self.read_bytes(4))[0]
+                    size -= 4 + len(f) + 1 + 4
+                    if size != file_length:
+                        raise errors.MessageError("Bad message size: {0}".format(size))
+                    if filename == '':
+                        # If there is no filename, then this is really RETURNREJECTED data, not a rejected file
+                        if file_length % 8 != 0:
+                            raise errors.MessageError("Bad RETURNREJECTED data size: {0}".format(file_length))
+                        data = self.read_bytes(file_length)
+                        message = messages.WriteFile(filename, file_length, data)
+                    else:
+                        # The rest of the message is read later with write_to_disk()
+                        message = messages.WriteFile(filename, file_length)
+                else:
+                    message = BackendMessage.from_type(type_, self.read_bytes(size - 4))
                 self._logger.debug('<= %s', message)
                 self.handle_asynchronous_message(message)
                 # handle transaction status
@@ -629,13 +672,22 @@ class Connection(object):
             raise errors.MessageError(msg)
 
     def read_bytes(self, n):
-        results = bytes()
-        while len(results) < n:
-            bytes_ = self._socket().recv(n - len(results))
-            if not bytes_:
+        if n == 1:
+            result = self._socket().recv(1)
+            if not result:
                 raise errors.ConnectionError("Connection closed by Vertica")
-            results += bytes_
-        return results
+            return result
+        else:
+            buf = bytearray(n)
+            view = memoryview(buf)
+            to_read = n
+            while to_read > 0:
+                received = self._socket().recv_into(view, to_read)
+                if received == 0:
+                    raise errors.ConnectionError("Connection closed by Vertica")
+                view = view[received:]
+                to_read -= received
+            return buf
 
     def send_GSS_response_and_receive_challenge(self, response):       
         # Send the GSS response data to the vertica server
