@@ -1,4 +1,4 @@
-# Copyright (c) 2018-2019 Micro Focus or one of its affiliates.
+# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
 # Copyright (c) 2018 Uber Technologies, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -37,8 +37,12 @@
 from __future__ import print_function, division, absolute_import
 
 import datetime
+import glob
 import inspect
 import re
+import sys
+import traceback
+from decimal import Decimal
 from io import IOBase
 from tempfile import NamedTemporaryFile, SpooledTemporaryFile, TemporaryFile
 from uuid import UUID
@@ -53,10 +57,10 @@ except ImportError:
 
 import six
 # noinspection PyUnresolvedReferences,PyCompatibility
-from builtins import str
-from six import binary_type, text_type, string_types, BytesIO, StringIO
+from six import binary_type, text_type, string_types, integer_types, BytesIO, StringIO
+from six.moves import zip
 
-from .. import errors
+from .. import errors, os_utils
 from ..compat import as_text
 from ..vertica import messages
 from ..vertica.column import Column
@@ -115,15 +119,16 @@ if six.PY2:
     # noinspection PyUnresolvedReferences
     file_type = file_type + (file,)
 
-NULL = "NULL"
 
-RE_NAME_BASE = u"[a-zA-Z_][\\w\\d\\$_]*"
+RE_NAME_BASE = u"[0-9a-zA-Z_][\\w\\d\\$_]*"
 RE_NAME = u'(("{0}")|({0}))'.format(RE_NAME_BASE)
 RE_BASIC_INSERT_STAT = (
     u"INSERT\\s+INTO\\s+(?P<target>({0}\\.)?{0})"
     u"\\s*\\(\\s*(?P<variables>{0}(\\s*,\\s*{0})*)\\s*\\)"
     u"\\s+VALUES\\s*\\(\\s*(?P<values>(.|\\s)*)\\s*\\)").format(RE_NAME)
 END_OF_RESULT_RESPONSES = (messages.CommandComplete, messages.PortalSuspended)
+END_OF_BATCH_RESPONSES = (messages.WriteFile, messages.EndOfBatchResponse)
+DEFAULT_BUFFER_SIZE = 131072
 
 
 class Cursor(object):
@@ -141,6 +146,7 @@ class Cursor(object):
         self.prepared_sql = None  # last statement been prepared
         self.prepared_name = "s0"
         self.error = None
+        self._sql_literal_adapters = {}
 
         #
         # dbapi attributes
@@ -171,17 +177,30 @@ class Cursor(object):
             self._close_prepared_statement()
         self._closed = True
 
-    def execute(self, operation, parameters=None, use_prepared_statements=None):
-        operation = as_text(operation)
-        self.operation = operation
-
+    def execute(self, operation, parameters=None, use_prepared_statements=None,
+                copy_stdin=None, buffer_size=DEFAULT_BUFFER_SIZE):
         if self.closed():
             raise errors.InterfaceError('Cursor is closed')
 
         self.flush_to_query_ready()
 
+        operation = as_text(operation)
+        self.operation = operation
+
         self.rowcount = -1
 
+        if copy_stdin is None:
+            self.copy_stdin_list = []
+        elif (isinstance(copy_stdin, list) and
+              all(callable(getattr(i, 'read', None)) for i in copy_stdin)):
+            self.copy_stdin_list = copy_stdin
+        elif callable(getattr(copy_stdin, 'read', None)):
+            self.copy_stdin_list = [copy_stdin]
+        else:
+            raise TypeError("Cursor.execute 'copy_stdin' parameter should be"
+                            " a file-like object or a list of file-like objects")
+        self.buffer_size = buffer_size   # For copy-local read and write
+
         use_prepared = bool(self.connection.options['use_prepared_statements']
                 if use_prepared_statements is None else use_prepared_statements)
         if use_prepared:
@@ -205,9 +224,6 @@ class Cursor(object):
         return self
 
     def executemany(self, operation, seq_of_parameters, use_prepared_statements=None):
-        operation = as_text(operation)
-        self.operation = operation
-
         if not isinstance(seq_of_parameters, (list, tuple)):
             raise TypeError("seq_of_parameters should be list/tuple")
 
@@ -215,6 +231,10 @@ class Cursor(object):
             raise errors.InterfaceError('Cursor is closed')
 
         self.flush_to_query_ready()
+
+        operation = as_text(operation)
+        self.operation = operation
+
         use_prepared = bool(self.connection.options['use_prepared_statements']
                 if use_prepared_statements is None else use_prepared_statements)
 
@@ -241,7 +261,7 @@ class Cursor(object):
 
                 values = as_text(m.group('values'))
                 values = ",".join([value.strip().strip('"') for value in values.split(",")])
-                seq_of_values = [self.format_operation_with_parameters(values, parameters, is_csv=True)
+                seq_of_values = [self.format_operation_with_parameters(values, parameters, is_copy_data=True)
                                  for parameters in seq_of_parameters]
                 data = "\n".join(seq_of_values)
 
@@ -276,6 +296,12 @@ class Cursor(object):
                 return None
             elif isinstance(self._message, messages.EmptyQueryResponse):
                 pass
+            elif isinstance(self._message, messages.VerifyFiles):
+                self._handle_copy_local_protocol()
+            elif isinstance(self._message, messages.EndOfBatchResponse):
+                pass
+            elif isinstance(self._message, messages.CopyDoneResponse):
+                pass
             elif isinstance(self._message, messages.ErrorResponse):
                 raise errors.QueryError.from_error_response(self._message, self.operation)
             else:
@@ -320,14 +346,19 @@ class Cursor(object):
             if isinstance(self._message, messages.RowDescription):
                 self.description = [Column(fd, self.unicode_error) for fd in self._message.fields]
                 self._message = self.connection.read_message()
+                if isinstance(self._message, messages.VerifyFiles):
+                    self._handle_copy_local_protocol()
+                self.rowcount = -1
                 return True
             elif isinstance(self._message, messages.BindComplete):
                 self._message = self.connection.read_message()
+                self.rowcount = -1
                 return True
             elif isinstance(self._message, messages.ReadyForQuery):
                 return False
             elif isinstance(self._message, END_OF_RESULT_RESPONSES):
                 # result of a DDL/transaction
+                self.rowcount = -1
                 return True
             elif isinstance(self._message, messages.ErrorResponse):
                 raise errors.QueryError.from_error_response(self._message, self.operation)
@@ -390,9 +421,14 @@ class Cursor(object):
         else:
             raise TypeError("Not valid type of data {0}".format(type(data)))
 
+        # TODO: check sql is a valid `COPY FROM STDIN` SQL statement
+
         self._logger.info(u'Execute COPY statement: [{}]'.format(sql))
+        # Execute a `COPY FROM STDIN` SQL statement
         self.connection.write(messages.Query(sql))
 
+        buffer_size = kwargs['buffer_size'] if 'buffer_size' in kwargs else DEFAULT_BUFFER_SIZE
+
         while True:
             message = self.connection.read_message()
 
@@ -401,17 +437,34 @@ class Cursor(object):
                 raise errors.QueryError.from_error_response(message, sql)
             elif isinstance(message, messages.ReadyForQuery):
                 break
-            elif isinstance(message, messages.CopyInResponse):
-                self.connection.write(messages.CopyStream(stream, **kwargs))
-                self.connection.write(messages.CopyDone())
             elif isinstance(message, messages.CommandComplete):
                 pass
+            elif isinstance(message, messages.CopyInResponse):
+                try:
+                    self._send_copy_data(stream, buffer_size)
+                except Exception as e:
+                    # COPY termination: report the cause of failure to the backend
+                    self.connection.write(messages.CopyFail(str(e)))
+                    self._logger.error(str(e))
+                    self.flush_to_query_ready()
+                    raise errors.DataError('Failed to send a COPY data stream: {}'.format(str(e)))
+
+                # Successful termination for COPY
+                self.connection.write(messages.CopyDone())
             else:
                 raise errors.MessageError('Unexpected message: {0}'.format(message))
 
         if self.error is not None:
             raise self.error
 
+    def object_to_sql_literal(self, py_obj):
+        return self.object_to_string(py_obj, False)
+
+    def register_sql_literal_adapter(self, obj_type, adapter_func):
+        if not callable(adapter_func):
+            raise TypeError("Cannot register this sql literal adapter. The adapter is not callable.")
+        self._sql_literal_adapters[obj_type] = adapter_func
+
     #############################################
     # internal
     #############################################
@@ -426,6 +479,9 @@ class Cursor(object):
             if isinstance(message, messages.ReadyForQuery):
                 self._message = message
                 break
+            elif isinstance(message, messages.VerifyFiles):
+                self._message = message
+                self._handle_copy_local_protocol()
 
     def flush_to_end_of_result(self):
         # if the last message isn't empty or END_OF_RESULT_RESPONSES,
@@ -453,16 +509,51 @@ class Cursor(object):
 
     def format_row_as_dict(self, row_data):
         return OrderedDict(
-            (self.description[idx].name, self.description[idx].convert(value))
-            for idx, value in enumerate(row_data.values)
+            (descr.name, descr.convert(value))
+            for descr, value in zip(self.description, row_data.values)
         )
 
     def format_row_as_array(self, row_data):
-        return [self.description[idx].convert(value)
-                for idx, value in enumerate(row_data.values)]
+        return [descr.convert(value)
+                for descr, value in zip(self.description, row_data.values)]
+
+    def object_to_string(self, py_obj, is_copy_data):
+        """Return the SQL representation of the object as a string"""
+        if type(py_obj) in self._sql_literal_adapters and not is_copy_data:
+            adapter = self._sql_literal_adapters[type(py_obj)]
+            result = adapter(py_obj)
+            if not isinstance(result, (string_types, bytes)):
+                raise TypeError("Unexpected return type of {} adapter: {}, expected a string type."
+                    .format(type(py_obj), type(result)))
+            return as_text(result)
+
+        if isinstance(py_obj, type(None)):
+            return '' if is_copy_data else 'NULL'
+        elif isinstance(py_obj, bool):
+            return str(py_obj)
+        elif isinstance(py_obj, (string_types, bytes)):
+            return self.format_quote(as_text(py_obj), is_copy_data)
+        elif isinstance(py_obj, (integer_types, float, Decimal)):
+            return str(py_obj)
+        elif isinstance(py_obj, tuple):  # tuple and namedtuple
+            elements = [None] * len(py_obj)
+            for i in range(len(py_obj)):
+                elements[i] = self.object_to_string(py_obj[i], is_copy_data)
+            return "(" + ",".join(elements) + ")"
+        elif isinstance(py_obj, (datetime.datetime, datetime.date, datetime.time, UUID)):
+            return self.format_quote(as_text(str(py_obj)), is_copy_data)
+        else:
+            if is_copy_data:
+                return str(py_obj)
+            else:
+                msg = ("Cannot convert {} type object to an SQL string. "
+                       "Please register a new adapter for this type via the "
+                       "Cursor.register_sql_literal_adapter() function."
+                       .format(type(py_obj)))
+                raise TypeError(msg)
 
     # noinspection PyArgumentList
-    def format_operation_with_parameters(self, operation, parameters, is_csv=False):
+    def format_operation_with_parameters(self, operation, parameters, is_copy_data=False):
         operation = as_text(operation)
 
         if isinstance(parameters, dict):
@@ -471,15 +562,7 @@ class Cursor(object):
                     key = str(key)
                 key = as_text(key)
 
-                if isinstance(param, (string_types, bytes)):
-                    param = self.format_quote(as_text(param), is_csv)
-                elif isinstance(param, (datetime.datetime, datetime.date, datetime.time, UUID)):
-                    param = self.format_quote(as_text(str(param)), is_csv)
-                elif param is None:
-                    param = '' if is_csv else NULL
-                else:
-                    param = str(param)
-                value = as_text(param)
+                value = self.object_to_string(param, is_copy_data)
 
                 # Using a regex with word boundary to correctly handle params with similar names
                 # such as :s and :start
@@ -489,16 +572,7 @@ class Cursor(object):
         elif isinstance(parameters, (tuple, list)):
             tlist = []
             for param in parameters:
-                if isinstance(param, (string_types, bytes)):
-                    param = self.format_quote(as_text(param), is_csv)
-                elif isinstance(param, (datetime.datetime, datetime.date, datetime.time, UUID)):
-                    param = self.format_quote(as_text(str(param)), is_csv)
-                elif param is None:
-                    param = '' if is_csv else NULL
-                else:
-                    param = str(param)
-                value = as_text(param)
-
+                value = self.object_to_string(param, is_copy_data)
                 tlist.append(value)
 
             operation = operation % tuple(tlist)
@@ -507,8 +581,8 @@ class Cursor(object):
 
         return operation
 
-    def format_quote(self, param, is_csv):
-        if is_csv:
+    def format_quote(self, param, is_copy_data):
+        if is_copy_data:
             return u'"{0}"'.format(re.escape(param))
         else:
             return u"'{0}'".format(param.replace(u"'", u"''"))
@@ -537,6 +611,147 @@ class Cursor(object):
             self._message = self.connection.read_message()
             if isinstance(self._message, messages.ErrorResponse):
                 raise errors.QueryError.from_error_response(self._message, query)
+            elif isinstance(self._message, messages.VerifyFiles):
+                self._handle_copy_local_protocol()
+
+    def _handle_copy_local_protocol(self):
+        if self.connection.options['disable_copy_local']:
+            msg = 'COPY LOCAL operation is disabled.'
+            self.connection.write(messages.CopyError(msg))
+            self.flush_to_query_ready()
+            raise errors.InterfaceError(msg)
+
+        # Extract info from VerifyFiles message
+        input_files = self._message.input_files
+        rejections_file = self._message.rejections_file
+        exceptions_file = self._message.exceptions_file
+
+        # Verify the file(s) present in the COPY FROM LOCAL statement are indeed accessible
+        self.valid_write_file_path = []
+        try:
+            # Check that the output files are writable
+            if rejections_file != '':
+                if rejections_file not in self.operation:
+                    raise errors.MessageError('Server requests for writing to'
+                        ' invalid rejected file path: {}'.format(rejections_file))
+                os_utils.check_file_writable(rejections_file)
+                self.valid_write_file_path.append(rejections_file)
+            if exceptions_file != '':
+                if exceptions_file not in self.operation:
+                    raise errors.MessageError('Server requests for writing to'
+                        ' invalid exceptions file path: {}'.format(exceptions_file))
+                os_utils.check_file_writable(exceptions_file)
+                self.valid_write_file_path.append(exceptions_file)
+
+            # Check that the input files are readable
+            self.valid_read_file_path = self._check_copy_local_files(input_files)
+
+            self.connection.write(messages.VerifiedFiles(self.valid_read_file_path))
+        except Exception as e:
+            tb = sys.exc_info()[2]
+            stk = traceback.extract_tb(tb, 1)
+            self.connection.write(messages.CopyError(str(e), stk[0]))
+            self.flush_to_query_ready()
+            raise
+
+        # Server should be ready to receive copy data from STDIN/files
+        self._message = self.connection.read_message()
+        if isinstance(self._message, messages.ErrorResponse):
+            raise errors.QueryError.from_error_response(self._message, self.operation)
+        elif not isinstance(self._message, (messages.CopyInResponse, messages.LoadFile)):
+            raise errors.MessageError('Unexpected COPY FROM LOCAL state: {}'.format(
+                                      type(self._message).__name__))
+        try:
+            if isinstance(self._message, messages.CopyInResponse):
+                self._logger.info('Sending STDIN data to server')
+                if len(self.copy_stdin_list) == 0:
+                    raise ValueError('No STDIN source to load. Please specify "copy_stdin" parameter in Cursor.execute()')
+                stdin = self.copy_stdin_list.pop(0)
+                self._send_copy_data(stdin, self.buffer_size)
+                self.connection.write(messages.EndOfBatchRequest())
+                self._read_copy_data_response(is_stdin_copy=True)
+            elif isinstance(self._message, messages.LoadFile):
+                while True:
+                    self._send_copy_file_data()
+                    if not self._read_copy_data_response():
+                        break
+        except Exception as e:
+            tb = sys.exc_info()[2]
+            stk = traceback.extract_tb(tb, 1)
+            self.connection.write(messages.CopyError(str(e), stk[0]))
+            self.flush_to_query_ready()
+            raise
+
+    def _check_copy_local_files(self, input_files):
+        # Return an empty list when the copy input is STDIN
+        if len(input_files) == 0:
+            return []
+
+        file_list = []
+        for file_pattern in input_files:
+            if file_pattern not in self.operation:
+                raise errors.MessageError('Server requests for loading invalid'
+                    ' file: {}, Query: {}'.format(file_pattern, self.operation))
+            # Expand the glob patterns
+            expanded_files = glob.glob(file_pattern)
+            if len(expanded_files) == 0:
+                raise OSError('{} does not exist'.format(file_pattern))
+            # Check file permissions
+            for f in expanded_files:
+                os_utils.check_file_readable(f)
+                file_list.append(f)
+        # Return a non-empty list when the copy input is FILE
+        # Note: Sending an empty list of files will make server kill the session.
+        return file_list
+
+    def _send_copy_data(self, stream, buffer_size):
+        # Send zero or more CopyData messages, forming a stream of input data
+        while True:
+            chunk = stream.read(buffer_size)
+            if not chunk:
+                break
+            self.connection.write(messages.CopyData(chunk, self.unicode_error))
+
+    def _send_copy_file_data(self):
+        filename = self._message.filename
+        self._logger.info('Sending {} data to server'.format(filename))
+
+        if filename not in self.valid_read_file_path:
+            raise errors.MessageError('Server requests for loading invalid'
+                    ' file: {}'.format(filename))
+
+        with open(filename, "rb") as f:
+            self._send_copy_data(f, self.buffer_size)
+        self.connection.write(messages.EndOfBatchRequest())
+
+    def _read_copy_data_response(self, is_stdin_copy=False):
+        """Return True if the server wants us to load more data, false if we are done"""
+        self._message = self.connection.read_expected_message(END_OF_BATCH_RESPONSES)
+        # Check for rejections during this load
+        while isinstance(self._message, messages.WriteFile):
+            if self._message.filename == '':
+                self._logger.info('COPY-LOCAL rejected row numbers: {}'.format(self._message.rejected_rows))
+            elif self._message.filename in self.valid_write_file_path:
+                self._message.write_to_disk(self.connection, self.buffer_size)
+            else:
+                raise errors.MessageError('Server requests for writing to'
+                    ' invalid file path: {}'.format(self._message.filename))
+            self._message = self.connection.read_expected_message(END_OF_BATCH_RESPONSES)
+
+        # For STDIN copy, there will be no incoming message until we send
+        # another EndOfBatchRequest or CopyDone
+        if is_stdin_copy:
+            self.connection.write(messages.CopyDone())  # End this copy
+            return False
+
+        # For file copy, peek the next message
+        self._message = self.connection.read_message()
+        if isinstance(self._message, messages.LoadFile):
+            # Indicate there are more local files to load
+            return True
+        elif not isinstance(self._message, messages.CopyDoneResponse):
+            raise errors.MessageError('Unexpected COPY-LOCAL message: {0}'.format(message))
+        return False
 
     def _error_handler(self, msg):
         self.connection.write(messages.Sync())
