@@ -1,4 +1,4 @@
-# Copyright (c) 2018-2019 Micro Focus or one of its affiliates.
+# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
 # Copyright (c) 2018 Uber Technologies, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -47,3 +47,7 @@ class Unknown(BackendMessage):
     @property
     def message_id(self):
         return self._message_id
+
+    def __str__(self):
+        return 'Unknown: message_id = {}, data = {}'.format(
+	           self._message_id, repr(self.data))
