@@ -1,4 +1,4 @@
-# Copyright (c) 2018-2019 Micro Focus or one of its affiliates.
+# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
 # Copyright (c) 2018 Uber Technologies, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -40,10 +40,11 @@ from .bind import Bind
 from .cancel_request import CancelRequest
 from .close import Close
 from .copy_data import CopyData
-from .copy_stream import CopyStream
 from .copy_done import CopyDone
+from .copy_error import CopyError
 from .copy_fail import CopyFail
 from .describe import Describe
+from .end_of_batch_request import EndOfBatchRequest
 from .execute import Execute
 from .flush import Flush
 from .load_balance_request import LoadBalanceRequest
@@ -54,7 +55,9 @@ from .ssl_request import SslRequest
 from .startup import Startup
 from .sync import Sync
 from .terminate import Terminate
+from .verified_files import VerifiedFiles
 
-__all__ = ['Bind', 'Query', 'CancelRequest', 'Close', 'CopyData', 'CopyDone', 'CopyFail',
-           'CopyStream', 'Describe', 'Execute', 'Flush', 'LoadBalanceRequest', 'Parse',
-           'Password', 'SslRequest', 'Startup', 'Sync', 'Terminate']
+__all__ = ['Bind', 'CancelRequest', 'Close', 'CopyData', 'CopyDone', 'CopyError',
+           'CopyFail', 'Describe', 'EndOfBatchRequest', 'Execute', 'Flush',
+           'LoadBalanceRequest', 'Parse', 'Password', 'Query', 'SslRequest', 'Startup',
+           'Sync', 'Terminate', 'VerifiedFiles']
