@@ -1,4 +1,4 @@
-# Copyright (c) 2018-2019 Micro Focus or one of its affiliates.
+# Copyright (c) 2018-2020 Micro Focus or one of its affiliates.
 # Copyright (c) 2018 Uber Technologies, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -33,6 +33,13 @@
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 # THE SOFTWARE.
 
+"""
+CopyFail message
+
+In the copy-in protocol, the frontend can terminate the cycle by sending a
+CopyFail message, which will cause the COPY SQL statement to fail with an error.
+"""
+
 from __future__ import print_function, division, absolute_import
 
 from struct import pack
@@ -45,7 +52,8 @@ class CopyFail(BulkFrontendMessage):
 
     def __init__(self, error_message):
         BulkFrontendMessage.__init__(self)
-        self._error_message = error_message
+        # An error message to report as the cause of failure
+        self._error_message = error_message.encode('utf-8')
 
     def read_bytes(self):
         bytes_ = pack('{0}sx'.format(len(self._error_message)), self._error_message)
